import { getCssAstEditor, getHtmlAstEditor, getJsAstEditor } from "@svelte-compose/ast-manipulation";
import {
    parseHtml,
    parsePostcss,
    parseScript,
    parseSvelteFile,
    serializeHtml,
    serializePostcss,
    serializeScript,
    serializeSvelteFile,
} from "@svelte-compose/ast-tooling";
import { fileExistsWorkspace, format, readFile, writeFile } from "./utils.js";

/**
 * @template {import("../composer/config").ArgType} Args
 * @param {import("../composer/config").FileTypes<Args>[]} files
 * @param {import("../composer/config").Workspace<Args>} workspace
 */
export async function createOrUpdateFiles(files, workspace) {
    for (const fileDetails of files) {
        if (fileDetails.condition && !fileDetails.condition(workspace)) {
            continue;
        }

        const exists = await fileExistsWorkspace(workspace, fileDetails.name(workspace));

        let content = "";
        if (!exists) {
            content = "";
        } else {
            content = await readFile(workspace, fileDetails.name(workspace));
        }

        if (fileDetails.contentType == "script") {
            content = handleScriptFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "text") {
            content = handleTextFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "svelte") {
            content = handleSvelteFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "json") {
            content = handleJsonFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "css") {
            content = handleCssFile(content, fileDetails, workspace);
        } else if (fileDetails.contentType == "html") {
            content = handleHtmlFile(content, fileDetails, workspace);
        }

        content = await format(workspace, fileDetails.name(workspace), content);
        await writeFile(workspace, fileDetails.name(workspace), content);
    }
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").HtmlFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleHtmlFile(content, fileDetails, workspace) {
    const ast = parseHtml(content);
    fileDetails.content({ ...getHtmlAstEditor(ast), ...workspace });
    content = serializeHtml(ast);
    return content;
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").CssFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleCssFile(content, fileDetails, workspace) {
    const ast = parsePostcss(content);
    fileDetails.content({ ...getCssAstEditor(ast), ...workspace });
    content = serializePostcss(ast);
    return content;
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").JsonFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleJsonFile(content, fileDetails, workspace) {
    const data = JSON.parse(content);
    fileDetails.content({ data, ...workspace });
    content = JSON.stringify(data, null, 4);
    return content;
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").SvelteFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleSvelteFile(content, fileDetails, workspace) {
    const { jsAst, htmlAst, cssAst } = parseSvelteFile(content);

    fileDetails.content({
        js: getJsAstEditor(jsAst),
        html: getHtmlAstEditor(htmlAst),
        css: getCssAstEditor(cssAst),
        ...workspace,
    });

    return serializeSvelteFile({ jsAst, htmlAst, cssAst });
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").TextFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleTextFile(content, fileDetails, workspace) {
    content = fileDetails.content({ content, ...workspace });
    return content;
}

/**
 *
 * @template {import("../composer/config").ArgType} Args
 * @param {string} content
 * @param {import("../composer/config").ScriptFileType<Args>} fileDetails
 * @param {import("../composer/config").Workspace<Args>} workspace
 * @returns {string}
 */
function handleScriptFile(content, fileDetails, workspace) {
    const ast = parseScript(content);

    fileDetails.content({
        ...getJsAstEditor(ast),
        ...workspace,
    });
    content = serializeScript(ast);
    return content;
}
